# -*- coding: utf-8 -*-

from unittest.mock import patch

from sklearn.ensemble import AdaBoostClassifier, RandomForestClassifier
from sklearn.linear_model import SGDClassifier
from sklearn.metrics import f1_score

from btb.benchmark.challenges.wind import WindABC, WindRFC, WindSGDC


@patch('btb.benchmark.challenges.challenge.OneHotEncoder')
@patch('btb.benchmark.challenges.challenge.MLChallenge.load_data')
def test_get_tunable_hyperparameters(mock_load_data, mokc_ohe):

    # setup
    mock_load_data.return_value = ('X', 'y')

    windabc = WindABC()
    windrfc = WindRFC()
    windsgdc = WindSGDC()

    # run
    abc_tunables = windabc.get_tunable_hyperparameters()
    rfc_tunables = windrfc.get_tunable_hyperparameters()
    sgdc_tunables = windsgdc.get_tunable_hyperparameters()

    # assert
    expected_abc_tunables = {
        "n_estimators": {
            "type": "int",
            "default": 50,
            "range": [
                1,
                500
            ]
        },
        "learning_rate": {
            "type": "float",
            "default": 1.0,
            "range": [
                1.0,
                10.0
            ]
        },
        "algorithm": {
            "type": "str",
            "default": "SAMME.R",
            "values": [
                "SAMME",
                "SAMME.R"
            ]
        }
    }

    expected_sgdc_tunables = {
        "loss": {
            "type": "str",
            "default": "hinge",
            "range": [
                "log",
                "hinge",
                "modified_huber",
                "squared_hinge",
                "perceptron",
                "squared_loss",
                "huber",
                "epsilon_insensitive",
                "squared_epsilon_insensitive"
            ]
        },
        "penalty": {
            "type": "str",
            "default": None,
            "values": [
                None,
                "l2",
                "l1",
                "elasticnet"
            ]
        },
        "alpha": {
            "type": "float",
            "default": 0.0001,
            "values": [
                0.0001,
                1
            ]
        },
        "max_iter": {
            "type": "int",
            "default": 1000,
            "values": [
                1,
                5000
            ]
        },
        "tol": {
            "type": "float",
            "default": 1e-3,
            "values": [
                1e-3,
                1
            ]
        },
        "shuffle": {
            "type": "bool",
            "default": True,
        }
    }

    expected_rfc_tunables = {
        "n_estimators": {
            "type": "int",
            "default": 10,
            "range": [
                1,
                500
            ]
        },
        "criterion": {
            "type": "str",
            "default": "gini",
            "values": [
                "entropy",
                "gini"
            ]
        },
        "max_features": {
            "type": "str",
            "default": None,
            "values": [
                None,
                "auto",
                "log2",
                "sqrt"
            ]
        },
        "min_samples_split": {
            "type": "int",
            "default": 2,
            "range": [
                2,
                100
            ]
        },
        "min_samples_leaf": {
            "type": "int",
            "default": 1,
            "range": [
                1,
                100
            ]
        },
        "min_weight_fraction_leaf": {
            "type": "float",
            "default": 0.0,
            "range": [
                0.0,
                0.5
            ]
        },
        "min_impurity_decrease": {
            "type": "float",
            "default": 0.0,
            "range": [
                0.0,
                1000.0
            ]
        }
    }

    assert expected_abc_tunables == abc_tunables
    assert expected_sgdc_tunables == sgdc_tunables
    assert expected_rfc_tunables == rfc_tunables


@patch('btb.benchmark.challenges.challenge.make_scorer')
@patch('btb.benchmark.challenges.challenge.OneHotEncoder')
@patch('btb.benchmark.challenges.challenge.MLChallenge.load_data')
def test_class_attributes(mock_load_data, mokc_ohe, mock_make_scorer):

    # setup
    mock_load_data.return_value = ('X', 'y')

    # run
    windabc = WindABC()
    windrfc = WindRFC()
    windsgdc = WindSGDC()

    # assert

    # ABC
    assert windabc.model == AdaBoostClassifier
    assert windabc.dataset == 'wind.csv'
    assert windabc.target_column == 'class'
    assert windabc.model_defaults == {'random_state': 0}
    assert windabc.make_binary
    assert windabc.metric == f1_score
    assert windabc.stratified

    # RFC
    assert windrfc.model == RandomForestClassifier
    assert windrfc.dataset == 'wind.csv'
    assert windrfc.target_column == 'class'
    assert windrfc.model_defaults == {'random_state': 0}
    assert windrfc.make_binary
    assert windrfc.metric == f1_score
    assert windrfc.stratified

    # SGDC
    assert windsgdc.model == SGDClassifier
    assert windsgdc.dataset == 'wind.csv'
    assert windsgdc.target_column == 'class'
    assert windsgdc.model_defaults == {'random_state': 0}
    assert windsgdc.make_binary
    assert windsgdc.metric == f1_score
    assert windsgdc.stratified



# -*- coding: utf-8 -*-

from unittest import TestCase

import numpy as np

from btb.benchmark.challenges.branin import Branin


class TestBranin(TestCase):

    def test___init__default(self):
        # run
        branin = Branin()

        # assert
        assert branin.a == 1
        assert branin.b == 5.1 / (4 * pow(np.pi, 2))
        assert branin.c == 5 / np.pi
        assert branin.r == 6
        assert branin.s == 10
        assert branin.t == 1 / (8 * np.pi)
        assert branin.min_x == -5.0
        assert branin.max_x == 10.0
        assert branin.min_y == 0.0
        assert branin.max_y == 15.0

    def test___init__custom(self):
        # run
        branin = Branin(a=1, b=2, c=3, r=4, s=5, t=6, min_x=4, max_x=5, min_y=6, max_y=7)

        # assert
        assert branin.a == 1
        assert branin.b == 2
        assert branin.c == 3
        assert branin.r == 4
        assert branin.s == 5
        assert branin.t == 6
        assert branin.min_x == 4
        assert branin.max_x == 5
        assert branin.min_y == 6
        assert branin.max_y == 7

    def test_get_tunable_hyperparameters(self):
        # run
        result = Branin().get_tunable_hyperparameters()

        # assert
        expected_result = {
            'x': {
                'type': 'float',
                'range': [-5.0, 10.0],
                'default': None
            },
            'y': {
                'type': 'float',
                'range': [0.0, 15.0],
                'default': None
            }
        }

        result == expected_result

    def test_get_tunable_custom_min_max(self):
        # run
        result = Branin(min_x=1, max_x=2, min_y=3, max_y=4).get_tunable_hyperparameters()

        # assert
        expected_result = {
            'x': {
                'type': 'float',
                'range': [1, 2],
                'default': None
            },
            'y': {
                'type': 'float',
                'range': [3, 4],
                'default': None
            }
        }

        result == expected_result

    def test_evaluate(self):
        # run
        result = Branin().evaluate(-np.pi, 12.275)
        result_2 = Branin().evaluate(np.pi, 2.275)
        result_3 = Branin().evaluate(10, 1)

        # assert
        assert result.round(6) == -0.397887
        assert result_2.round(6) == -0.397887
        assert result_3.round(6) == -5.954976



# -*- coding: utf-8 -*-

from unittest import TestCase

from btb.benchmark.challenges.rosenbrock import Rosenbrock


class TestRosenbrock(TestCase):

    def test___init__default(self):
        # run
        rosenbrock = Rosenbrock()

        # assert
        assert rosenbrock.a == 1
        assert rosenbrock.b == 100
        assert rosenbrock.min_x == -50
        assert rosenbrock.max_x == 50
        assert rosenbrock.min_y == -50
        assert rosenbrock.max_y == 50

    def test___init__custom(self):
        # run
        rosenbrock = Rosenbrock(a=2, b=3, min_x=4, max_x=5, min_y=6, max_y=7)

        # assert
        assert rosenbrock.a == 2
        assert rosenbrock.b == 3
        assert rosenbrock.min_x == 4
        assert rosenbrock.max_x == 5
        assert rosenbrock.min_y == 6
        assert rosenbrock.max_y == 7

    def test_get_tunable_hyperparameters(self):
        # run
        result = Rosenbrock().get_tunable_hyperparameters()

        # assert
        expected_result = {
            'x': {
                'type': 'int',
                'range': [-50, 50],
                'default': None
            },
            'y': {
                'type': 'int',
                'range': [-50, 50],
                'default': None
            }
        }

        assert result == expected_result

    def test_get_tunable_hyperparameters_custom_min_max(self):
        # run
        result = Rosenbrock(min_x=1, max_x=2, min_y=3, max_y=4).get_tunable_hyperparameters()

        # assert
        expected_result = {
            'x': {
                'type': 'int',
                'range': [1, 2],
                'default': None
            },
            'y': {
                'type': 'int',
                'range': [3, 4],
                'default': None
            }
        }

        assert result == expected_result

    def test_evaluate(self):
        # run
        result = Rosenbrock().evaluate(1, 2)
        result_2 = Rosenbrock().evaluate(1, 1)

        # assert
        assert result == -100
        assert result_2 == 0



# -*- coding: utf-8 -*-
import numpy as np

from btb.benchmark import benchmark
from btb.benchmark.challenges import Rosenbrock
from btb.benchmark.tuners.btb import make_tuning_function
from btb.tuning import GPTuner


def test_benchmark_rosenbrock():
    # run
    candidate = make_tuning_function(GPTuner)
    df = benchmark(candidate, challenges=Rosenbrock(), iterations=1)

    # Assert
    np.testing.assert_equal(df.columns.values, ['Rosenbrock()', 'Mean', 'Std'])
    np.testing.assert_equal(df.index.values, ['tuning_function'])
    np.testing.assert_equal(df.dtypes.values, [np.int, np.float, np.float])



from sklearn.ensemble import AdaBoostRegressor, BaggingRegressor, RandomForestRegressor
from sklearn.metrics import r2_score

from btb.benchmark.challenges.challenge import MLChallenge


class BostonRFR(MLChallenge):
    """
    This dataset contains information collected by the U.S Census Service
    concerning housing in the area of Boston. It was obtained from the
    StatLib archive http://lib.stat.cmu.edu/datasets/boston

    The prediction task is to determine the price of a house based on the
    location and other features that the dataset provides.
    """

    # DATSET
    DATASET = 'boston.csv'
    TARGET_COLUMN = 'medv'
    STRATIFIED = False

    # CROSS VALIDATE / SCORER
    METRIC = r2_score
    ENCODE = True
    MAKE_BINARY = False

    # MODEL
    MODEL = RandomForestRegressor
    MODEL_DEFAULTS = {'random_state': 0}
    TUNABLE_HYPERPARAMETERS = {
        "n_estimators": {
            "type": "int",
            "default": 10,
            "range": [
                1,
                500
            ]
        },
        "criterion": {
            "type": "str",
            "default": "mse",
            "values": [
                "mse",
                "mae"
            ]
        },
        "max_features": {
            "type": "str",
            "default": "auto",
            "range": [
                None,
                "auto",
                "log2",
                "sqrt"
            ]
        },
        "min_samples_split": {
            "type": "int",
            "default": 2,
            "range": [
                2,
                1000
            ]
        },
        "min_samples_leaf": {
            "type": "int",
            "default": 1,
            "range": [
                1,
                1000
            ]
        },
        "min_weight_fraction_leaf": {
            "type": "float",
            "default": 0.0,
            "range": [
                0.0,
                0.5
            ]
        },
        "min_impurity_decrease": {
            "type": "float",
            "default": 0.0,
            "range": [
                0.0,
                10.0
            ]
        }
    }


class BostonABR(BostonRFR):
    # MODEL
    MODEL = AdaBoostRegressor
    MODEL_DEFAULTS = {'random_state': 0}
    TUNABLE_HYPERPARAMETERS = {
        "n_estimators": {
            "type": "int",
            "default": 50,
            "range": [
                1,
                500
            ]
        },
        "learning_rate": {
            "type": "float",
            "default": 1.0,
            "range": [
                1.0,
                10.0
            ]
        },
        "loss": {
            "type": "str",
            "default": "linear",
            "values": [
                "linear",
                "square",
                "exponential"
            ]
        }
    }


class BostonBR(BostonRFR):
    # MODEL
    MODEL = BaggingRegressor
    MODEL_DEFAULTS = {'bootstrap': True, 'random_state': 0}
    TUNABLE_HYPERPARAMETERS = {
        "n_estimators": {
            "type": "int",
            "default": 10,
            "range": [
                1,
                500
            ]
        },
        "max_samples": {
            "type": "int",
            "default": 1,
            "range": [
                1,
                100
            ]
        },
    }



# -*- coding: utf-8 -*-

from unittest import TestCase

from btb.benchmark.challenges.bohachevsky import Bohachevsky


class TestBohachevsky(TestCase):

    def test___init__default(self):
        # run
        bohachevsky = Bohachevsky()

        # assert
        assert bohachevsky.min_x == -100
        assert bohachevsky.max_x == 100
        assert bohachevsky.min_y == -100
        assert bohachevsky.max_y == 100

    def test___init__custom(self):
        # run
        bohachevsky = Bohachevsky(min_x=4, max_x=5, min_y=6, max_y=7)

        # assert
        assert bohachevsky.min_x == 4
        assert bohachevsky.max_x == 5
        assert bohachevsky.min_y == 6
        assert bohachevsky.max_y == 7

    def test_get_tunable_hyperparameters(self):
        # run
        result = Bohachevsky().get_tunable_hyperparameters()

        # assert
        expected_result = {
            'x': {
                'type': 'int',
                'range': [-100, 100],
                'default': None
            },
            'y': {
                'type': 'int',
                'range': [-100, 100],
                'default': None
            }
        }

        assert result == expected_result

    def test_get_tunable_hyperparameters_custom_min_max(self):
        # run
        result = Bohachevsky(min_x=1, max_x=2, min_y=3, max_y=4).get_tunable_hyperparameters()

        # assert
        expected_result = {
            'x': {
                'type': 'int',
                'range': [1, 2],
                'default': None
            },
            'y': {
                'type': 'int',
                'range': [3, 4],
                'default': None
            }
        }

        assert result == expected_result

    def test_evaluate(self):
        # run
        result = Bohachevsky().evaluate(1, 2)
        result_2 = Bohachevsky().evaluate(0, 0)

        # assert
        assert result == -9.6
        assert result_2 == 0



# -*- coding: utf-8 -*-

"""Package where the Challenge class is defined."""

import inspect
import os
from abc import ABCMeta, abstractmethod
from copy import deepcopy
from urllib.parse import urljoin

import pandas as pd
from sklearn.metrics import make_scorer
from sklearn.model_selection import KFold, StratifiedKFold, cross_val_score
from sklearn.preprocessing import OneHotEncoder

BTB_DATA_URL = 'https://btb-data.s3.amazonaws.com/'


def _get_dataset_url(name):

    if not name.endswith('.gz'):
        name = name + '.gz'

    return urljoin(BTB_DATA_URL, name)


class Challenge(metaclass=ABCMeta):
    """Challenge class.

    The Challenge class represents a single ``challenge`` that can be used for benchmark.
    """

    @abstractmethod
    def get_tunable_hyperparameters(self):
        """Return a dictionary with hyperparameters to be tuned."""
        pass

    @abstractmethod
    def evaluate(self, *args, **kwargs):
        """Perform evaluation for the given ``arguments``.

        This method will score a result with a given configuration, then return the score obtained
        for those ``arguments``.
        """
        pass

    def __repr__(self):
        args = inspect.getargspec(self.__init__)
        keys = args.args[1:]
        defaults = dict(zip(keys, args.defaults))
        instanced = {key: getattr(self, key) for key in keys}

        if defaults == instanced:
            return '{}()'.format(self.__class__.__name__)

        else:
            args = ', '.join(
                '{}={}'.format(key, value)
                for key, value in instanced.items()
            )

            return '{}({})'.format(self.__class__.__name__, args)


class MLChallenge(Challenge):
    """Machine Learning Challenge class.

    The MLChallenge class rerpresents a single ``machine learning challenge`` that can be used for
    benchmark.

    Args:
        model (class):
            Class of a machine learning estimator.

        dataset (str):
            Name or path to a dataset. If it's a name it will try to read it from
            https://btb-data.s3.amazonaws.com/

        target_column (str):
            Name of the target column in the dataset.

        encode (bool):
            Either or not to encode the dataset using ``sklearn.preprocessing.OneHotEncoder``.

        model_defaults (dict):
            Dictionary with default keyword args for the model instantiation.

        make_binary (bool):
            Either or not to make the target column binary.

        tunable_hyperparameters (dict):
            Dictionary representing the tunable hyperparameters for the challenge.

        metric (callable):
            Metric function. If ``None``, then the estimator's metric function
            will be used in case there is otherwise the default that ``cross_val_score`` function
            offers will be used.
    """

    def load_data(self):
        """Load ``X`` and ``y`` over which to perform fit and evaluate."""
        if os.path.isdir(self.dataset):
            X = pd.read_csv(self.dataset)

        else:
            url = _get_dataset_url(self.dataset)
            X = pd.read_csv(url, compression='gzip')

        y = X.pop(self.target_column)

        if self.make_binary:
            y = y.iloc[0] == y

        return X, y

    def __init__(self, model=None, dataset=None, target_column=None,
                 encode=None, tunable_hyperparameters=None, metric=None,
                 model_defaults=None, make_binary=None, stratified=None,
                 cv_splits=5, cv_random_state=42, cv_shuffle=True):

        self.model = model or self.MODEL
        self.dataset = dataset or self.DATASET
        self.target_column = target_column or self.TARGET_COLUMN
        self.model_defaults = model_defaults or self.MODEL_DEFAULTS
        self.make_binary = make_binary or self.MAKE_BINARY
        self.tunable_hyperparameters = tunable_hyperparameters or self.TUNABLE_HYPERPARAMETERS

        if metric:
            self.metric = metric
        else:
            # Allow to either write a metric method or assign a METRIC function
            self.metric = getattr(self, 'metric', self.__class__.METRIC)

        self.stratified = self.STRATIFIED if stratified is None else stratified
        self.X, self.y = self.load_data()

        self.encode = self.ENCODE if encode is None else encode
        self.scorer = make_scorer(self.metric)

        if self.stratified:
            self.cv = StratifiedKFold(
                shuffle=cv_shuffle,
                n_splits=cv_splits,
                random_state=cv_random_state
            )
        else:
            self.cv = KFold(
                shuffle=cv_shuffle,
                n_splits=cv_splits,
                random_state=cv_random_state
            )

        if self.encode:
            ohe = OneHotEncoder()
            self.X = ohe.fit_transform(self.X)

    def get_tunable_hyperparameters(self):
        return deepcopy(self.tunable_hyperparameters)

    def evaluate(self, **hyperparams):
        """Apply cross validation to hyperparameter combination.

        Args:
            hyperparams (dict):
                A combination of ``self.tunable_hyperparams``.

        Returns:
            score (float):
                Returns the ``mean`` cross validated score.
        """
        hyperparams.update((self.model_defaults or {}))
        model = self.model(**hyperparams)
        return cross_val_score(model, self.X, self.y, cv=self.cv, scoring=self.scorer).mean()

    def __repr__(self):
        return '{}()'.format(self.__class__.__name__)



# -*- coding: utf-8 -*-

from sklearn.ensemble import AdaBoostClassifier, RandomForestClassifier
from sklearn.linear_model import SGDClassifier
from sklearn.metrics import f1_score

from btb.benchmark.challenges.challenge import MLChallenge


class CensusRFC(MLChallenge):
    """
    The Census challenge is based on the `Census Income Dataset`. The extraction
    was done by Barry Becker from the 1994 Census Database. The prediction task
    is to determine whether a person makes over 50.000 USD a year.

    This dataset has been obtained from https://archive.ics.uci.edu/ml/datasets/census+income
    """

    # DATASET
    DATASET = 'census.csv'
    TARGET_COLUMN = 'income'

    # CROSS VALIDATE / SCORER
    METRIC = f1_score
    ENCODE = True
    MAKE_BINARY = True
    STRATIFIED = True

    # MODEL
    MODEL = RandomForestClassifier
    MODEL_DEFAULTS = {'random_state': 0}
    TUNABLE_HYPERPARAMETERS = {
        "n_estimators": {
            "type": "int",
            "default": 10,
            "range": [
                1,
                500
            ]
        },
        "criterion": {
            "type": "str",
            "default": "gini",
            "values": [
                "entropy",
                "gini"
            ]
        },
        "max_features": {
            "type": "str",
            "default": None,
            "values": [
                None,
                "auto",
                "log2",
                "sqrt"
            ]
        },
        "min_samples_split": {
            "type": "int",
            "default": 2,
            "range": [
                2,
                100
            ]
        },
        "min_samples_leaf": {
            "type": "int",
            "default": 1,
            "range": [
                1,
                100
            ]
        },
        "min_weight_fraction_leaf": {
            "type": "float",
            "default": 0.0,
            "range": [
                0.0,
                0.5
            ]
        },
        "min_impurity_decrease": {
            "type": "float",
            "default": 0.0,
            "range": [
                0.0,
                1000.0
            ]
        },
    }


class CensusSGDC(CensusRFC):
    # MODEL
    MODEL = SGDClassifier
    TUNABLE_HYPERPARAMETERS = {
        "loss": {
            "type": "str",
            "default": "hinge",
            "range": [
                "log",
                "hinge",
                "modified_huber",
                "squared_hinge",
                "perceptron",
                "squared_loss",
                "huber",
                "epsilon_insensitive",
                "squared_epsilon_insensitive"
            ]
        },
        "penalty": {
            "type": "str",
            "default": None,
            "values": [
                None,
                "l2",
                "l1",
                "elasticnet"
            ]
        },
        "alpha": {
            "type": "float",
            "default": 0.0001,
            "values": [
                0.0001,
                1
            ]
        },
        "max_iter": {
            "type": "int",
            "default": 1000,
            "values": [
                1,
                5000
            ]
        },
        "tol": {
            "type": "float",
            "default": 1e-3,
            "values": [
                1e-3,
                1
            ]
        },
        "shuffle": {
            "type": "bool",
            "default": True,
        },
    }


class CensusABC(CensusRFC):
    # MODEL
    MODEL = AdaBoostClassifier
    TUNABLE_HYPERPARAMETERS = {
        "n_estimators": {
            "type": "int",
            "default": 50,
            "range": [
                1,
                500
            ]
        },
        "learning_rate": {
            "type": "float",
            "default": 1.0,
            "range": [
                1.0,
                10.0
            ]
        },
        "algorithm": {
            "type": "str",
            "default": "SAMME.R",
            "values": [
                "SAMME",
                "SAMME.R"
            ]
        }
    }



# -*- coding: utf-8 -*-

import numpy as np

from btb.tuning.tunable import Tunable


def tune(tuner, scoring_function, iterations):
    best_score = -np.inf

    for _ in range(iterations):
        proposal = tuner.propose()
        score = scoring_function(**proposal)
        tuner.record(proposal, score)
        best_score = max(score, best_score)

    return best_score


def make_tuning_function(tuner_class, **tuner_kwargs):
    def tuning_function(scoring_function, tunable_hyperparameters, iterations):
        tunable = Tunable.from_dict(tunable_hyperparameters)
        tuner = tuner_class(tunable, **tuner_kwargs)
        return tune(tuner, scoring_function, iterations)

    return tuning_function



# -*- coding: utf-8 -*-

from unittest import TestCase
from unittest.mock import MagicMock, patch

import pandas as pd

from btb.benchmark import benchmark, evaluate_candidate


class TestBenchmark(TestCase):

    def test_evaluate_candidate_challenge_is_instance(self):
        # setup
        candidate = MagicMock()
        challenge = MagicMock()
        challenge.__str__.return_value = 'test_challenge'

        # run
        result = evaluate_candidate('test_candidate', candidate, challenge, 10)

        # assert
        challenge.get_tunable_hyperparameters.assert_called_once_with()
        candidate.assert_called_once_with(
            challenge.evaluate,
            challenge.get_tunable_hyperparameters.return_value,
            10
        )

        expected_result = [
            {
                'challenge': 'test_challenge',
                'candidate': 'test_candidate',
                'score': candidate.return_value
            }
        ]

        assert result == expected_result

    @patch('btb.benchmark.evaluate_candidate')
    def test_benchmark_challenges_callable(self, mock_evaluate_candidate):
        # setup
        mock_evaluate_candidate.return_value = [{
            'challenge': 'test_challenge',
            'candidate': 'test_candidate',
            'score': 1.0
        }]

        candidate = MagicMock(__name__='test_candidate')
        challenge = MagicMock()

        # run
        result = benchmark(candidate, challenges=challenge)

        # assert
        expected_result = pd.DataFrame({
            'test_challenge': [1.0],
            'Mean': [1.0],
            'Std': [0.0],
        })

        expected_result.index = ['test_candidate']

        mock_evaluate_candidate.assert_called_once_with(
            'test_candidate',
            candidate,
            [challenge],
            1000
        )

        pd.testing.assert_frame_equal(
            result.sort_index(axis=1),
            expected_result.sort_index(axis=1),
        )

    @patch('btb.benchmark.evaluate_candidate')
    def test_benchmark_challenges_tuple(self, mock_evaluate_candidate):
        # setup
        mock_evaluate_candidate.side_effect = [
            [{'challenge': 'test_challenge', 'candidate': 'candidate_a', 'score': 1.0}],
            [{'challenge': 'test_challenge', 'candidate': 'candidate_b', 'score': 1.0}],
        ]

        candidate_a = MagicMock(return_value=0.1, __name__='candidate_a')
        candidate_b = MagicMock(return_value=0.1, __name__='candidate_b')
        candidates = (candidate_a, candidate_b)
        challenge = MagicMock()

        # run
        result = benchmark(candidates, challenges=challenge)

        # assert
        expected_result = pd.DataFrame({
            'test_challenge': [1.0, 1.0],
            'Mean': [1.0, 1.0],
            'Std': [0.0, 0.0],
        })

        expected_result.index = ['candidate_a', 'candidate_b']

        pd.testing.assert_frame_equal(
            result.sort_index(axis=1),
            expected_result.sort_index(axis=1),
        )

    def test_benchmark_candidates_not_dict_not_callable(self):
        # setup
        with self.assertRaises(TypeError):
            benchmark(1)



# -*- coding: utf-8 -*-

from unittest import TestCase
from unittest.mock import MagicMock, patch

from btb.benchmark.challenges.challenge import MLChallenge


class TestMLChallenge(TestCase):

    @patch('btb.benchmark.challenges.challenge.make_scorer')
    @patch('btb.benchmark.challenges.challenge.StratifiedKFold')
    @patch('btb.benchmark.challenges.challenge.MLChallenge.load_data')
    def test___init__stratified(self, mock_load_data, mock_strfkfold, mock_make_scorer):

        # setup
        mock_load_data.return_value = (1, 2)
        mock_make_scorer.return_value = 'test_scorer'
        mock_strfkfold.return_value = 'cv'

        # run
        instance = MLChallenge(
            model='test',
            dataset='any',
            target_column='test_column',
            encode=False,
            tunable_hyperparameters='test_hp',
            metric='f1_score',
            model_defaults='any',
            make_binary=True,
            stratified=True
        )

        # assert
        assert instance.model == 'test'
        assert instance.dataset == 'any'
        assert instance.model == 'test'
        assert instance.dataset == 'any'
        assert instance.target_column == 'test_column'
        assert not instance.encode
        assert instance.tunable_hyperparameters == 'test_hp'
        assert instance.scorer == 'test_scorer'
        assert instance.model_defaults == 'any'
        assert instance.make_binary
        assert instance.cv == 'cv'

        mock_strfkfold.assert_called_once_with(shuffle=True, n_splits=5, random_state=42)

    @patch('btb.benchmark.challenges.challenge.make_scorer')
    @patch('btb.benchmark.challenges.challenge.KFold')
    @patch('btb.benchmark.challenges.challenge.MLChallenge.load_data')
    def test___init__kfold(self, mock_load_data, mock_kfold, mock_make_scorer):

        # setup
        mock_load_data.return_value = (1, 2)
        mock_make_scorer.return_value = 'test_scorer'
        mock_kfold.return_value = 'kfold'

        # run
        instance = MLChallenge(
            model='test',
            dataset='any',
            target_column='test_column',
            encode=False,
            tunable_hyperparameters='test_hp',
            metric='f1_score',
            model_defaults='any',
            make_binary=True,
            stratified=False
        )

        # assert
        assert instance.model == 'test'
        assert instance.dataset == 'any'
        assert instance.model == 'test'
        assert instance.dataset == 'any'
        assert instance.target_column == 'test_column'
        assert not instance.encode
        assert instance.tunable_hyperparameters == 'test_hp'
        assert instance.scorer == 'test_scorer'
        assert instance.model_defaults == 'any'
        assert instance.make_binary
        assert instance.cv == 'kfold'

        mock_kfold.assert_called_once_with(shuffle=True, n_splits=5, random_state=42)

    @patch('btb.benchmark.challenges.challenge.make_scorer')
    @patch('btb.benchmark.challenges.challenge.StratifiedKFold')
    @patch('btb.benchmark.challenges.challenge.OneHotEncoder')
    def test___init__encode(self, mock_ohe, mock_strfkfold, mock_make_scorer):

        # setup
        instance = MagicMock()
        instance.load_data.return_value = ('X', 'y')
        instance.encode = True
        instance.DATASET = 'dataset'

        # run
        MLChallenge.__init__(instance, metric='test', encode=True)

        # assert
        mock_ohe.return_value.fit_transform.assert_called_once_with('X')

    def test_get_tunable_hyperparameters(self):
        # setup
        instance = MagicMock()
        instance.tunable_hyperparameters = {'test': 'hyperparam'}

        # run
        result = MLChallenge.get_tunable_hyperparameters(instance)

        # assert
        assert result == {'test': 'hyperparam'}

    @patch('btb.benchmark.challenges.challenge.cross_val_score')
    def test_evaluate(self, mock_crossval):
        # setup
        mock_crossval.return_value.mean.return_value = 1
        instance = MagicMock()
        instance.scorer = 'scoring'
        instance.cv = 'cv'

        # run
        result = MLChallenge.evaluate(instance)

        # assert
        assert result == 1

        mock_crossval(instance.model.return_value, 'X', 'y', cv='cv', scoring='scoring')



# -*- coding: utf-8 -*-

from unittest.mock import patch

from sklearn.ensemble import AdaBoostRegressor, BaggingRegressor, RandomForestRegressor
from sklearn.metrics import r2_score

from btb.benchmark.challenges.boston import BostonABR, BostonBR, BostonRFR


@patch('btb.benchmark.challenges.challenge.OneHotEncoder')
@patch('btb.benchmark.challenges.challenge.MLChallenge.load_data')
def test_get_tunable_hyperparameters(mock_load_data, mock_ohe):

    # setup
    mock_load_data.return_value = ('X', 'y')

    bostonabr = BostonABR()
    bostonbr = BostonBR()
    bostonrfr = BostonRFR()

    # run
    abr_tunable_hyperparameters = bostonabr.get_tunable_hyperparameters()
    br_tunable_hyperparameters = bostonbr.get_tunable_hyperparameters()
    rfr_tunable_hyperparameters = bostonrfr.get_tunable_hyperparameters()

    # assert
    expected_abr_hyperparameters = {
        "n_estimators": {
            "type": "int",
            "default": 50,
            "range": [
                1,
                500
            ]
        },
        "learning_rate": {
            "type": "float",
            "default": 1.0,
            "range": [
                1.0,
                10.0
            ]
        },
        "loss": {
            "type": "str",
            "default": "linear",
            "values": [
                "linear",
                "square",
                "exponential"
            ]
        }
    }

    expected_br_hyperparameters = {
        "n_estimators": {
            "type": "int",
            "default": 10,
            "range": [
                1,
                500
            ]
        },
        "max_samples": {
            "type": "int",
            "default": 1,
            "range": [
                1,
                100
            ]
        },
    }

    expected_rfr_hyperparameters = {
        "n_estimators": {
            "type": "int",
            "default": 10,
            "range": [
                1,
                500
            ]
        },
        "criterion": {
            "type": "str",
            "default": "mse",
            "values": [
                "mse",
                "mae"
            ]
        },
        "max_features": {
            "type": "str",
            "default": "auto",
            "range": [
                None,
                "auto",
                "log2",
                "sqrt"
            ]
        },
        "min_samples_split": {
            "type": "int",
            "default": 2,
            "range": [
                2,
                1000
            ]
        },
        "min_samples_leaf": {
            "type": "int",
            "default": 1,
            "range": [
                1,
                1000
            ]
        },
        "min_weight_fraction_leaf": {
            "type": "float",
            "default": 0.0,
            "range": [
                0.0,
                0.5
            ]
        },
        "min_impurity_decrease": {
            "type": "float",
            "default": 0.0,
            "range": [
                0.0,
                10.0
            ]
        }
    }

    assert expected_abr_hyperparameters == abr_tunable_hyperparameters
    assert expected_br_hyperparameters == br_tunable_hyperparameters
    assert expected_rfr_hyperparameters == rfr_tunable_hyperparameters


@patch('btb.benchmark.challenges.challenge.make_scorer')
@patch('btb.benchmark.challenges.challenge.OneHotEncoder')
@patch('btb.benchmark.challenges.challenge.MLChallenge.load_data')
def test_class_attributes(mock_load_data, mokc_ohe, mock_make_scorer):

    # setup
    mock_load_data.return_value = ('X', 'y')

    # run
    bostonabr = BostonABR()
    bostonbr = BostonBR()
    bostonrfr = BostonRFR()

    # assert

    # ABR
    assert bostonabr.model == AdaBoostRegressor
    assert bostonabr.dataset == 'boston.csv'
    assert bostonabr.target_column == 'medv'
    assert bostonabr.model_defaults == {'random_state': 0}
    assert not bostonabr.make_binary
    assert bostonabr.metric == r2_score
    assert not bostonabr.stratified

    # BR
    assert bostonbr.model == BaggingRegressor
    assert bostonbr.dataset == 'boston.csv'
    assert bostonbr.target_column == 'medv'
    assert bostonbr.model_defaults == {'bootstrap': True, 'random_state': 0}
    assert not bostonbr.make_binary
    assert bostonbr.metric == r2_score
    assert not bostonbr.stratified

    # RFR
    assert bostonrfr.model == RandomForestRegressor
    assert bostonrfr.dataset == 'boston.csv'
    assert bostonrfr.target_column == 'medv'
    assert bostonrfr.model_defaults == {'random_state': 0}
    assert not bostonrfr.make_binary
    assert bostonrfr.metric == r2_score
    assert not bostonrfr.stratified



# -*- coding: utf-8 -*-

"""Top level where all the challenges are imported."""

from btb.benchmark.challenges.bohachevsky import Bohachevsky
from btb.benchmark.challenges.boston import BostonABR, BostonBR, BostonRFR
from btb.benchmark.challenges.branin import Branin
from btb.benchmark.challenges.census import CensusABC, CensusRFC, CensusSGDC
from btb.benchmark.challenges.rosenbrock import Rosenbrock
from btb.benchmark.challenges.wind import WindABC, WindRFC, WindSGDC

__all__ = (
    'Bohachevsky',
    'BostonABR',
    'BostonBR',
    'BostonRFR',
    'Branin',
    'CensusABC',
    'CensusRFC',
    'CensusSGDC',
    'Rosenbrock',
    'WindABC',
    'WindRFC',
    'WindSGDC',
)



# -*- coding: utf-8 -*-

import pandas as pd

from btb.benchmark.challenges.bohachevsky import Bohachevsky
from btb.benchmark.challenges.boston import BostonABR, BostonBR, BostonRFR
from btb.benchmark.challenges.branin import Branin
from btb.benchmark.challenges.census import CensusABC, CensusRFC, CensusSGDC
from btb.benchmark.challenges.rosenbrock import Rosenbrock
from btb.benchmark.challenges.wind import WindABC, WindRFC, WindSGDC

DEFAULT_CHALLENGES = [
    # Simple
    Bohachevsky,
    Branin,
    Rosenbrock,

    # ML
    BostonABR,
    BostonBR,
    BostonRFR,
    CensusABC,
    CensusRFC,
    CensusSGDC,
    WindABC,
    WindRFC,
    WindSGDC
]


def evaluate_candidate(name, candidate, challenges, iterations):
    candidate_result = []

    if not isinstance(challenges, list):
        challenges = [challenges]

    for challenge in challenges:
        tunable_hyperparameters = challenge.get_tunable_hyperparameters()
        score = candidate(challenge.evaluate, tunable_hyperparameters, iterations)
        result = {
            'challenge': str(challenge),
            'candidate': name,
            'score': score,
        }

        candidate_result.append(result)

    return candidate_result


def benchmark(candidates, challenges=None, iterations=1000):
    """Benchmark function.

    This benchmark function iterates over a collection of ``challenges`` and executes a
    ``tuner_function`` for each one of the ``challenges`` for a given amount of iterations.

    Args:
        candidates (callable, list, tuple or dict):
            Python callable function, list of callable functions, tuple with callable functions or
            dictionary with the ``name`` of the function as ``key`` and the callable function that
            returns the best score for a given ``scorer``. This function must have three arguments:

                * scorer (function):
                    A function that performs scoring over params.

                * tunable (btb.tuning.Tunable):
                    A ``Tunable`` instance used to instantiate a tuner.

                * iterations (int):
                    Number of tuning iterations to perform.

        challenges (single challenge or list):
            A single ``challenge`` or a list of ``chalenges``. This challenges must inherit
            from ``btb.challenges.challenge.Challenge``.
        iterations (int):
            Amount of iterations to perform for the ``tuner_function``.

    Returns:
        pandas.DataFrame:
            A ``pandas.DataFrame`` with the obtained scores for the given challenges is being
            returned.
    """
    if challenges is None:
        challenges = [challenge_class() for challenge_class in DEFAULT_CHALLENGES]

    if callable(candidates):
        candidates = {candidates.__name__: candidates}

    elif isinstance(candidates, (list, tuple)):
        candidates = {candidate.__name__: candidate for candidate in candidates}

    elif not isinstance(candidates, dict):
        raise TypeError(
            'Candidates can only be a callable, list of callables, tuple of callables or dict.')

    if not isinstance(challenges, list):
        challenges = [challenges]

    results = []

    for name, candidate in candidates.items():
        result = evaluate_candidate(name, candidate, challenges, iterations)
        results.extend(result)

    df = pd.DataFrame.from_records(results)
    df = df.pivot(index='candidate', columns='challenge', values='score')

    del df.columns.name
    del df.index.name

    df['Mean'] = df.mean(axis=1)
    df['Std'] = df.std(axis=1)

    return df
